# file: $RCSfile: jwp-astro.tcl,v $
# rcsid: $Id: jwp-astro.tcl,v 1.4 2006/08/07 21:24:53 jwp Exp $
# Copyright Jeffrey W Percival
# ######################################################################
# Space Astronomy Laboratory
# University of Wisconsin
# 1150 University Avenue
# Madison, WI 53706 USA
# ######################################################################
# Do not use this software without permission.
# Do not use this software without attribution.
# Do not remove or alter any of the lines above.
# ######################################################################

# ######################################################################
# $RCSfile: jwp-astro.tcl,v $ - Tcl implementation of JWP's libastro.a
# ######################################################################
#
# ######################################################################
# A note for the procedures in this file:
# This file implements a subset of the
# Telescope Pointing Machine (TPM) software by Jeffrey W Percival.
# TPM does the usual astronomical coordinate transformations
# to milli-arcsecond accuracy,
# and matches the performance of SLALIB and NOVAS.
# See http://www.sal.wisc.edu/~jwp/astro/tpm/tpm.html
#
# An important deviation from my TPM C-code
# is that TPM implements the astronomical target
# as a 6-D true-length position and space-motion velocity vector,
# while WHOMP is more concerned
# with the projection of the Virgo chip onto the sky.
# This nuance calls for attitude matrices, not vectors,
# so we can embody roll (z-axis rotation) as well.
#
# So my TPM components (e.g. the gal2equ subroutine)
# operate here on attitude matrices, not TPM-style state vectors.
#
# ######################################################################
#

# from tcllib (http://www.tcl.tk/software/tcllib/)
source constants.tcl
source linalg.tcl
package require math::constants
package require math::linearalgebra

# pick up our support files
source jwp-times.tcl
source jwp-vec.tcl
source jwp-lis.tcl

namespace eval astro {

variable GAL_RA  192.25;
variable GAL_DEC 27.4;
variable GAL_LON 33.0;

# Compute the mean obliquity of the ecliptic for the given equinox.
# From Exp. Supp., 1992, p. 114
proc obliquity { tdt } {

    set tdt [expr double($tdt)];	# enforce type

    set T [expr ($tdt - $times::J2000) / $times::CJ];

    set obl [expr 84381.448 + $T * (-46.8150 + ($T * (-0.00059 + ($T * 0.001813))))];

    set obl [times::as2r $obl];

    return $obl;
}

# Produce the conversion matrix for Galactic to FK4 Equatorial.
# The galactic pole is at (ra,dec) = (192.25, 27.4) degrees,
# and the longitude of the ascending node of the galactic plane
# on the equator is 33 degrees.
# The transformation is Rz(-ra) * Ry(-(90-dec)) * Rz(-(90-lon))
proc gal2equ_m { } {
    variable GAL_RA;
    variable GAL_DEC;
    variable GAL_LON;

    set a1 [times::d2r [expr -1 * (90 - $GAL_LON)]]
    set a2 [times::d2r [expr -1 * (90 - $GAL_DEC)]]
    set a3 [times::d2r [expr -1 * ($GAL_RA)]]

    set r1 [vec::m3Rz $a1];
    set r2 [vec::m3Ry $a2];
    set r3 [vec::m3Rz $a3];

    set gm [math::linearalgebra::mkIdentity 3];
    set gm [math::linearalgebra::matmul $r1 $gm];
    set gm [math::linearalgebra::matmul $r2 $gm];
    set gm [math::linearalgebra::matmul $r3 $gm];

    return $gm;
}

# Convert from FK4 Galactic to FK4 Equatorial.
proc gal2equ { m } {

    set gm [gal2equ_m];
    set gm [math::linearalgebra::transpose $gm];

    set m [math::linearalgebra::matmul $m $gm];

    return $m;
}

# Produce the FK4->FK5 transformation matrix.
proc fk425_m { } {
    set m [math::linearalgebra::mkIdentity 3];
    math::linearalgebra::setrow m 0 { 9.999256781650602e-01 -1.118206108957399e-02 -4.857947723285060e-03 };
    math::linearalgebra::setrow m 1 { 1.118206101909165e-02  9.999374784319467e-01 -2.717645143240717e-05 };
    math::linearalgebra::setrow m 2 { 4.857947885514070e-03 -2.714743630320534e-05  9.999881997331131e-01 };
    return $m;
}

# Convert an FK4 position to FK5.
# Ignore elliptical aberration (eterms) for now.
proc fk425 { m } {

    set fm [fk425_m];
    set fm [math::linearalgebra::transpose $fm];

    set m [math::linearalgebra::matmul $m $fm];

    return $m;
}

# Produce the conversion matrix for IAU 1980 Ecliptic to FK5 Equatorial.
proc ecl2equ_m { utc } {

    set obl [obliquity $utc];

    set em [vec::m3Rx [expr -1 * $obl]];

    return $em;
}

# Convert a position from IAU 1980 Ecliptic to FK5 Equatorial.
proc ecl2equ { m utc } {

    set em [ecl2equ_m $utc];
    set em [math::linearalgebra::transpose $em];

    set m [math::linearalgebra::matmul $m $em];

    return $m;
}

#
# this section gives expressions for the precession angles
# used to produce an FK5 precession matrix.
#

proc zeta { j1 j2 } {

    set T [expr ($j1 -$times::J2000) / $times::CJ];
    set t [expr ($j2 - $j1) / $times::CJ];
    set zeta (0.017998);
    set zeta [expr $zeta * $t];
    set zeta [expr $zeta + (0.30188 + ($T * (-0.000344)))];
    set zeta [expr $zeta * $t];
    set zeta [expr $zeta + (2306.2181 + ($T * (1.39656 + ($T * (-0.000139)))))];
    set zeta [expr $zeta * $t];

    # convert to radians
    set zeta [times::as2r $zeta];

    return $zeta;
}

# we ignore this for now
proc zetadot { j1 j2 } {
    return 0;
}

proc theta { j1 j2 } {

    set T [expr ($j1 -$times::J2000) / $times::CJ];
    set t [expr ($j2 - $j1) / $times::CJ];
    set theta (-0.041833);
    set theta [expr $theta * $t];
    set theta [expr $theta + (-0.42665 + ($T * (-0.000217)))];
    set theta [expr $theta * $t];
    set theta [expr $theta + (2004.3109 + ($T * (-0.85330 + ($T * (-0.000217)))))];
    set theta [expr $theta * $t];

    # convert to radians

    set theta [times::as2r $theta];

    return $theta;
}

# we ignore this for now
proc thetadot { j1 j2 } {
    return 0;
}

proc zee { j1 j2 } {

    set T [expr ($j1 -$times::J2000) / $times::CJ];
    set t [expr ($j2 - $j1) / $times::CJ];
    set zee (0.018203);
    set zee [expr $zee * $t];
    set zee [expr $zee + (1.09468 + ($T * (0.000066)))];
    set zee [expr $zee * $t];
    set zee [expr $zee + (2306.2181 + ($T * (1.39656 + ($T * (-0.000139)))))];
    set zee [expr $zee * $t];

    # convert to radians

    set zee [times::as2r $zee];

    return $zee;
}

# we ignore this for now
proc zeedot { j1 j2 } {
    return 0;
}

#
# End of precession angle section.
#

# Produce a precession matrix.
proc precess_m { j1 j2 } {

    set a1 [zeta $j1 $j2];	set a1 [expr -1 * $a1];
    set a2 [theta $j1 $j2];
    set a3 [zee $j1 $j2];	set a3 [expr -1 * $a3];

    set r1 [vec::m3Rz $a1];
    set r2 [vec::m3Ry $a2];
    set r3 [vec::m3Rz $a3];

    set pm [math::linearalgebra::mkIdentity 3];
    set pm [math::linearalgebra::matmul $r1 $pm];
    set pm [math::linearalgebra::matmul $r2 $pm];
    set pm [math::linearalgebra::matmul $r3 $pm];

    return $pm;
}

# Precess an FK5 attitude matrix.
proc precess { j1 j2 m } {

    set pm [precess_m $j1 $j2];
    set pm [math::linearalgebra::transpose $pm];

    set m [math::linearalgebra::matmul $m $pm];

    return $m;
}

proc test { } {

    puts stdout "astro::test:";

    set j1 $times::J2000;
    set j2 [times::ymd2j {2006 0 0 {0 0 0}}];

    set m [lis::sky2matrix 0 0 0];
    set m [gal2equ $m];
    set radec [lis::matrix2sky $m];
    set alpha [lindex $radec 0];
    set delta [lindex $radec 1];
    puts stdout "astro::test: gal2equ: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta]";

    set m [lis::sky2matrix 0 0 0];
    set m [fk425 $m];
    set radec [lis::matrix2sky $m];
    set alpha [lindex $radec 0];
    set delta [lindex $radec 1];
    puts stdout "astro::test:   fk425: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta]";

    set m [lis::sky2matrix [times::h2r 6] [times::d2r 90] 0];
    set m [ecl2equ $m $j2];
    set radec [lis::matrix2sky $m];
    set alpha [lindex $radec 0];
    set delta [lindex $radec 1];
    puts stdout "astro::test: ecl2equ: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta]";

    set m [lis::sky2matrix 0 0 0];
    set m [precess $j1 $j2 $m];
    set radec [lis::matrix2sky $m];
    set alpha [lindex $radec 0];
    set delta [lindex $radec 1];
    puts stdout "astro::test: precess: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta]";

    set m [lis::sky2matrix 0 0 0];
    set pm [precess_m $j1 $j2];
    set pminv [math::linearalgebra::transpose $pm];
    set m [math::linearalgebra::matmul $m $pminv];
    set radec [lis::matrix2sky $m];
    set alpha [lindex $radec 0];
    set delta [lindex $radec 1];
    puts stdout "astro::test: precess: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta]";

    # move along the galactic equator,
    # expressing the attitude in FK5.
    # see that the roll is correct.
    foreach i {-2 -1 0 1 2} {
	set m [lis::sky2matrix 0 0 0];
	#puts stdout "astro::test: m:\n[math::linearalgebra::show $m "%12.8f"]";
	set r [vec::m3Ry $i];
	#puts stdout "astro::test: r:\n[math::linearalgebra::show $r "%12.8f"]";
	set m [math::linearalgebra::matmul $r $m];
	#puts stdout "astro::test: m:\n[math::linearalgebra::show $m "%12.8f"]";
	set gm [gal2equ_m];
	#puts stdout "astro::test: gm:\n[math::linearalgebra::show $gm "%12.8f"]";
	set gminv [math::linearalgebra::transpose $gm];
	#puts stdout "astro::test: gminv:\n[math::linearalgebra::show $gminv "%12.8f"]";
	set m [math::linearalgebra::matmul $m $gminv];
	set radec [lis::matrix2sky $m];
	set alpha [lindex $radec 0];
	set delta [lindex $radec 1];
	set roll [lindex $radec 2];
	puts stdout "astro::test: gal2equ: alpha: [times::fmt_alpha $alpha] delta: [times::fmt_delta $delta] roll: [times::fmt_r $roll]";
    }

}

# end of namespace
}
