# file: $RCSfile: jwp-lis.tcl,v $
# rcsid: $Id: jwp-lis.tcl,v 1.3 2006/08/07 21:26:27 jwp Exp $
# Copyright Jeffrey W Percival
# ######################################################################
# Space Astronomy Laboratory
# University of Wisconsin
# 1150 University Avenue
# Madison, WI 53706 USA
# ######################################################################
# Do not use this software without permission.
# Do not use this software without attribution.
# Do not remove or alter any of the lines above.
# ######################################################################

# ######################################################################
# $RCSfile: jwp-lis.tcl,v $ - Tcl implementation of JWP's liblis.a
# ######################################################################

# ######################################################################
# This file implements some of the mighty Star Tracker 5000's attitude code.
# ######################################################################

# from tcllib (http://www.tcl.tk/software/tcllib/)
source constants.tcl
source linalg.tcl
package require math::constants
package require math::linearalgebra

# pick up our support files
source jwp-vec.tcl
source jwp-times.tcl

namespace eval lis {

# Convert celestial coordinates to Euler angles.
# Notes:
# 1. the celestial coordinates can be Equatorial, Ecliptic, or Galactic
# 2. we define the angles thus:
#    alpha: the longitudinal angle (e.g. RA)
#    delta: the latitudinal angle (e.g. Declination)
#    roll: the final roll about the z-axis
# 3. we assume the usual 3-1-3 form of attitude matrix
proc sky2euler { alpha delta roll } {
    set alpha [expr double($alpha)];	# enforce type
    set delta [expr double($delta)];	# enforce type
    set roll [expr double($roll)];	# enforce type

    set phi [expr $alpha + $math::constants::pi/2];
    set theta [expr $math::constants::pi/2 - $delta];
    set psi $roll;

    return [list $phi $theta $psi];
}

# Convert the Euler angles into an attitude matrix.
# We use the angles in a 3-1-3 (z-x-z) order.
proc euler2matrix { phi theta psi } {
    set phi [expr double($phi)];	# enforce type
    set theta [expr double($theta)];	# enforce type
    set psi [expr double($psi)];	# enforce type

    # set up our three rotations
    set r1 [vec::m3Rz $phi];
    set r2 [vec::m3Rx $theta];
    set r3 [vec::m3Rz $psi];

    # compute the attitude matrix
    set m [math::linearalgebra::mkIdentity 3];
    set m [math::linearalgebra::matmul $r1 $m];
    set m [math::linearalgebra::matmul $r2 $m];
    set m [math::linearalgebra::matmul $r3 $m];

    return $m;
}

# Convert an attitude matrix into Euler angles.
# We assume a 3-1-3 sequence of rotations.
# Matrix indices are 0,1,2 but my variable names use 1,2,3.
# This is so they match my C and Java code.
proc matrix2euler { m } {

    set m33 [math::linearalgebra::getelem $m 2 2];
    set cos_theta [expr abs($m33)];

    # we have an ambiguity if sin(theta) = 0; test using a 1 arcsec threshold
    set epsilon [times::as2r 1];
    set cos_epsilon [expr cos($epsilon)];
    if {$cos_theta > $cos_epsilon} {
	# theta is a multiple of pi; resolve multiple with cos(theta)
	if {$m33 > 0} {
	    set theta 0;
	} else {
	    set theta $math::constants::pi;
	}

	# We can only find the sum or difference of psi and phi,
	# depending on whether theta is an even or odd multiple of pi.
	# Choose phi=0, and assign all the angle to psi.
	set phi 0;
	set m21 [math::linearalgebra::getelem $m 1 0];
	set m11 [math::linearalgebra::getelem $m 0 0];
	set psi [expr atan2(-1*$m21, $m11)];
    } else {
	set m31 [math::linearalgebra::getelem $m 2 0];
	set m32 [math::linearalgebra::getelem $m 2 1];
	set phi [expr atan2($m31, -1*$m32)];

	set theta [expr acos($m33)];

	set m13 [math::linearalgebra::getelem $m 0 2];
	set m23 [math::linearalgebra::getelem $m 1 2];
	set psi [expr atan2($m13, $m23)];

	# resolve the sign ambiguity in theta;
	# use the sin(psi)*sin(theta) element
	if {$m13 < 0} {
	    # either sin(psi) or sin(theta) (but not both) must be negative
	    if {sin($psi) > 0} {
		# it must have been theta
		set theta [expr -1 * $theta];
	    }
	} else {
	    # they must share the sign
	    if {sin($psi) < 0} {
		set theta [expr -1 * $theta];
	    }
	}
    }
    return [list $phi $theta $psi];
}

# Convert Euler angles into celestial coordinates.
proc euler2sky { phi theta psi } {
    set phi [expr double($phi)];	# enforce type
    set theta [expr double($theta)];	# enforce type
    set psi [expr double($psi)];	# enforce type

    set alpha [expr $phi - $math::constants::pi/2];
    set delta [expr $math::constants::pi/2 - $theta];
    set roll $psi;

    return [list $alpha $delta $roll];
}

# Convert celestial coordinates to an attitude matrix.
# We hide the Euler angle step from the user.
proc sky2matrix { alpha delta roll } {
    set alpha [expr double($alpha)];	# enforce type
    set delta [expr double($delta)];	# enforce type
    set roll [expr double($roll)];	# enforce type

    set euler [sky2euler $alpha $delta $roll];
    set matrix [eval euler2matrix $euler];

    return $matrix;
}

# Convert an attitude matrix to celestial coordinates.
# We hide the Euler angle step from the user.
proc matrix2sky { matrix } {

    set euler [matrix2euler $matrix];
    set sky [eval euler2sky $euler];

    return $sky;
}

proc test { } {

    puts stdout "lis::test:";

    set alpha 1;
    set delta 0.5;
    set roll 2;

    puts stdout "lis::test: alpha: $alpha delta: $delta roll: $roll";

    set euler [sky2euler $alpha $delta $roll];
    puts stdout "lis::test: euler: $euler";

    set matrix [eval euler2matrix $euler];
    puts stdout "lis::test: matrix:\n[math::linearalgebra::show $matrix "%12.8f"]";

    set euler [matrix2euler $matrix];
    puts stdout "lis::test: euler: $euler";

    set sky [eval euler2sky $euler];
    puts stdout "lis::test: sky: $sky";

    set matrix [sky2matrix $alpha $delta $roll];
    puts stdout "lis::test: matrix:\n[math::linearalgebra::show $matrix "%12.8f"]";

    set sky [matrix2sky $matrix];
    puts stdout "lis::test: sky: $sky";
}

# end of namespace
}
