# file: $RCSfile: jwp-times.tcl,v $
# rcsid: $Id: jwp-times.tcl,v 1.5 2010/04/27 20:17:14 jwp Exp $
# Copyright Jeffrey W Percival
# ######################################################################
# Space Astronomy Laboratory
# University of Wisconsin
# 1150 University Avenue
# Madison, WI 53706 USA
# ######################################################################
# Do not use this software without permission.
# Do not use this software without attribution.
# Do not remove or alter any of the lines above.
# ######################################################################

# ######################################################################
# $RCSfile: jwp-times.tcl,v $ - Tcl implementation of JWP's libtimes.a
# ######################################################################

# ######################################################################
# A note for the procedures in this file:
# we treat angles and times as both scalars and as 3-element lists;
# the lists represent the usual deg/min/sec or hour/min/sec triplets.
#
# Normalization: we allow each element of a triplet to range over any value;
# for example, the minutes field can be < 0 or > 60.
# We provide normalization routines
# to put things into proper "human" format for display.
# ######################################################################

# from tcllib (http://www.tcl.tk/software/tcllib/)
source constants.tcl
package require math::constants

namespace eval times {

######################################
# set up some astronomical constants #
######################################

# the JD of the modified JD system
variable MJD_0 2400000.5;

# the JD of B1950.0
variable B1950 2433282.42345905;

# the JD of 1984.0, the magic FK4/FK5 conversion point
variable J1984 2445700.5;

# the JD of J2000.0
variable J2000 2451545.0;

# the tropical century as 1900.0
variable CB 36524.21987817305;

# the Julian century
variable CJ 36525.0;

####################################################
# create some simple angle/time utility procedures #
# these are implemented as C macros in Unix-land.  #
####################################################

# convert radians to arcseconds
proc r2as { r } {
    set d [r2d $r];
    set as [expr $d*3600.0];
    return $as;
}

# convert arcseconds to radians
proc as2r { as } {
    set d [expr $as/3600.0];
    set r [d2r $d];
    return $r;
}

##################################
# now define the main procedures #
##################################

# compute the julian day number of the given besselian year
proc byear2j { byear } {
    variable B1950;
    variable CB;

    set byear [expr double($byear)];	# enforce type

    set j [expr ($B1950 + ($byear - 1950.0) * ($CB/100))];

    return $j;
}

# normalize scalar degrees into the range (-360, 360)
proc d2d { d } {
    set d [expr double($d)];	# enforce type

    if {$d <= -360} {
	set d [expr $d + ceil($d / -360.0) * 360.0];
    }

    if {$d >= 360} {
	set d [expr $d - floor($d / 360.0) * 360.0];
    }

    return $d;
}

# convert scalar degrees to dms degrees
proc d2dms { d } {
    set d [expr double($d)];	# enforce type

    set dms [list $d 0.0 0.0];

    return $dms;
}

# convert scalar degrees to scalar hours
proc d2h { d } {
    set d [expr double($d)];	# enforce type

    set h [expr $d / 15.0];

    return $h;
}

# convert scalar degrees to hms hours
proc d2hms { d } {
    set d [expr double($d)];	# enforce type

    set h [d2h $d];
    set hms [h2hms $h];

    return $hms;
}

# convert scalar degrees to scalar radians
proc d2r { d } {
    set d [expr double($d)];	# enforce type

    set r [expr $d*($math::constants::pi/180)];

    return $r;
}

# convert dms degrees to scalar degrees
proc dms2d { dms } {

    set d 0.0;
    set d [expr $d + [lindex $dms 0]];
	if {$d < 0.0} {
		set d [expr $d - [lindex $dms 1] / 60.0];
		set d [expr $d - [lindex $dms 2] / 3600.0];
	} else {
		set d [expr $d + [lindex $dms 1] / 60.0];
		set d [expr $d + [lindex $dms 2] / 3600.0];
	}

    return $d;
}

# normalize dms degrees
proc dms2dms { dms } {

    # convert to decimal degrees
    set d [dms2d $dms];
    lset dms 0 [expr floor($d)];

    # promote the minutes part
    set m [expr ($d - floor($d)) * 60.0];
    lset dms 1 [expr floor($m)];

    # promote the seconds part
    set s [expr ($m - floor($m)) * 60.0];
    lset dms 2 $s;

    # normalize the degrees
    lset dms 0 [d2d [lindex $dms 0]];

    return $dms;
}

# convert dms degrees to scalar hours
proc dms2h { dms } {

    set d [dms2d $dms];
    set h [d2h $d];

    return $h;
}

# convert dms degrees to hms hours
proc dms2hms { dms } {

    set hms {0.0 0.0 0.0};
    lset hms 0 [expr [lindex $dms 0] / 15.0];
    lset hms 1 [expr [lindex $dms 1] / 15.0];
    lset hms 2 [expr [lindex $dms 2] / 15.0];

    return $hms;
}

# convert dms degrees to scalar radians
proc dms2r { dms } {

    set d [dms2d $dms];
    set r [d2r $d];

    return $r;
}

# format scalar radians as time from 0 to 24 hours
proc fmt_alpha { alpha } {
    set alpha [expr double($alpha)];	# enforce type

    set twopi [expr 2*$math::constants::pi];

    if {$alpha <= 0.0} {
	set alpha [expr $alpha + ceil($alpha / (-1*$twopi)) * $twopi];
    }
    if {$alpha >= $twopi} {
	set alpha [expr $alpha - floor($alpha / (-1*$twopi)) * $twopi];
    }

    set hms [r2hms $alpha];

    return [fmt_hms $hms];
}

# format scalar degrees
proc fmt_d { d } {
    set d [expr double($d)];	# enforce type

    set sign "+";
    if {$d < 0.0} {
	set sign "-";
	set d [expr abs($d)];
    }
    set deg [expr int(floor($d))];
    set m [expr ($d - floor($d)) * 60.0];
    set min [expr int(floor($m))];
    set s [expr ($m - floor($m)) * 60.0];

    set ipart [expr int($s)];
    set fpart [expr int(1e3 * ($s - $ipart))];

    set p [format "%s%02d:%02d:%02d.%03d" $sign $deg $min $ipart $fpart];

    return $p;
}

# format scalar radians as degrees in the range (-90, 90)
proc fmt_delta { delta } {
    set delta [expr double($delta)];	# enforce type

    set twopi [expr 2*$math::constants::pi];

    if {$delta <= -1*$math::constants::pi} {
    	set delta [expr $delta + ceil($delta / (-1*$twopi)) * $twopi];
    }
    if {$delta > $math::constants::pi} {
    	set delta [expr $delta - floor($delta / ($twopi)) * $twopi];
    }
    if {$delta > $math::constants::pi/2} {
    	set delta [expr $math::constants::pi - $delta];
    }
    if {$delta < -1*$math::constants::pi/2} {
    	set delta [expr -1*$math::constants::pi - $delta];
    }

    set dms [r2dms $delta];

    return [fmt_dms $dms];
}

# format dms degrees
proc fmt_dms { dms } {

    set d [dms2d $dms];

    return [fmt_d $d];
}

# format scalar hours
proc fmt_h { h } {
    set h [expr double($h)];	# enforce type

    set sign " ";	# no plus sign for hours
    if {$h < 0.0} {
	set sign "-";
	set h [expr abs($h)];
    }

    set hrs [expr int(floor($h))];
    set m [expr ($h - floor($h)) * 60.0];
    set min [expr int(floor($m))];
    set s [expr ($m - floor($m)) * 60.0];

    set ipart [expr int($s)];
    set fpart [expr int(1e3 * ($s - $ipart))];

    set p [format "%s%02d:%02d:%02d.%03d" $sign $hrs $min $ipart $fpart];

    return $p;
}

# format hms hours
proc fmt_hms { hms } {

    set h [hms2h $hms];

    return [fmt_h $h];
}

# format scalar julian date
proc fmt_j { j } {
    set j [expr double($j)];	# enforce type

    set d [expr floor($j)];
    set h [expr ($j - $d) * 24.0];
    set p [format "%3d %s" [expr int($d)] [fmt_h $h]];

    return $p;
}

# format jd time
proc fmt_jd { jd } {

    set j [jd2j $jd];

    return [fmt_j $j];
}

# format scalar radians
proc fmt_r { r } {
    set r [expr double($r)];	# enforce type

    set d [r2d $r];

    return [fmt_d $d];
}

# format scalar years
proc fmt_y { y } {
    set y [expr double($y)];	# enforce type

    set ymd [y2ymd $y];

    return [fmt_ymd $ymd];
}

# format ymd time
proc fmt_ymd { ymd } {

    set DOW(0) "Sun";
    set DOW(1) "Mon";
    set DOW(2) "Tue";
    set DOW(3) "Wed";
    set DOW(4) "Thu";
    set DOW(5) "Fri";
    set DOW(6) "Sat";

    set MOY(0) "Jan";
    set MOY(1) "Feb";
    set MOY(2) "Mar";
    set MOY(3) "Apr";
    set MOY(4) "May";
    set MOY(5) "Jun";
    set MOY(6) "Jul";
    set MOY(7) "Aug";
    set MOY(8) "Sep";
    set MOY(9) "Oct";
    set MOY(10) "Nov";
    set MOY(11) "Dec";

    # normalize the time
    set ymd [ymd2ymd $ymd];

    # get the julian date
    set j [ymd2j $ymd];

    # break up the ymd
    set y [expr int([lindex $ymd 0])];
    set m [expr int([lindex $ymd 1])];
    set dd [lindex $ymd 2];
    set hh [lindex $ymd 3 0];
    set mm [lindex $ymd 3 1];
    set ss [lindex $ymd 3 2];

    # get the day of the week
    set today [j2dow $j];

    # get the seconds
    set ipart [expr int($ss)];
    set fpart [expr int(1e3 * ($ss - $ipart))];

    if {$y < 0} {
	set y [expr 1 - $y];
    }

    set p [format "%3.3s %3.3s %2d %02d:%02d:%02d.%03d %4d" \
    	$DOW([expr $today]) \
	$MOY([expr $m-1])  \
	[expr int($dd)]  \
	[expr int($hh)] \
	[expr int($mm)] \
	$ipart $fpart \
	$y];

    return $p;

}

# map gregorian calendar date into scalar julian date
proc gcal2j { y m d } {
    set y [expr int($y)];	# enforce type
    set m [expr int($m)];	# enforce type
    set d [expr int($d)];	# enforce type

    # normalize the month
    if {$m < 1} {
	set y [expr $y - (1-$m)/12];
	set m [expr ($m % 12) + 12];
    }
    if {$m > 12} {
	set y [expr $y + ($m-1)/12];
	set m [expr (($m-1) % 12) + 1];
    }

    set j 0;

    #
    # In C, -7/12 is 0; in Tcl, it's -1!  Argh.
    # We will cobble up the C behavior
    #
    if {($m - 14) < 0} {
	set M [expr (14 - $m) / 12];
	set M [expr -1 * $M];
    } else {
	set M [expr ($m - 14) / 12];
    }

    set j [expr $j + ((1461 * ($y + 4800 + $M)) / 4)];
    set j [expr $j + ((367 * ($m - 2 - 12*($M))) / 12)]
    set j [expr $j - ((3 * (($y + 4900 + $M) / 100)) / 4)];
    set j [expr $j + $d];
    set j [expr $j - 32075];

    return [expr double($j)];
}

# convert scalar hours to scalar degrees
proc h2d { h } {
    set h [expr double($h)];	# enforce type

    set d [expr $h * 15.0];

    return $d;
}

# convert scalar hours to dms degrees
proc h2dms { h } {
    set h [expr double($h)];	# enforce type

    set d [h2d $h];
    set dms [d2dms $d];

    return $dms;
}

# normalize scalar hours into the range (0, 24)
proc h2h { h } {
    set h [expr double($h)];	# enforce type

    if {$h <= 0.0} {
	set h [expr $h + ceil($h / -24.0) * 24.0];
    }
    if {$h >= 24.0} {
	set h [expr $h - floor($h / 24.0) * 24.0];
    }

    return $h;
}

# convert scalar hours to hms hours
proc h2hms { h } {
    set h [expr double($h)];	# enforce type

    set hms [list $h 0.0 0.0];

    return $hms;
}

# convert scalar hours to scalar radians
proc h2r { h } {
    set h [expr double($h)];	# enforce type

    set r [expr $h*($math::constants::pi/12)];

    return $r;
}

# convert hms hours to scalar degrees
proc hms2d { hms } {

    set h [hms2h $hms];
    set d [h2d $h];

    return $d;
}

# convert hms hours into dms degrees
proc hms2dms { hms } {

    set dms {0.0 0.0 0.0};
    lset dms 0 [expr [lindex $hms 0] * 15.0];
    lset dms 1 [expr [lindex $hms 1] * 15.0];
    lset dms 2 [expr [lindex $hms 2] * 15.0];

    return $dms;
}

# convert hms hours to scalar hours
proc hms2h { hms } {

    set h 0.0;
    set h [expr $h + [lindex $hms 0]];
    set h [expr $h + [lindex $hms 1] / 60.0];
    set h [expr $h + [lindex $hms 2] / 3600.0];

    return $h;
}

# normalize hms hours
proc hms2hms { hms } {

    # convert to decimal hours
    set h [hms2h $hms];
    lset hms 0 [expr floor($h)];

    # promote the minutes part
    set m [expr ($h - floor($h)) * 60.0];
    lset hms 1 [expr floor($m)];

    # promote the seconds part
    set s [expr ($m - floor($m)) * 60.0];
    lset hms 2 $s;

    # we do not normalize the hours,
    # as we did degrees in dms2dms,
    # as time is not periodic like angle.

    return $hms;
}

# convert hms hours to scalar radians
proc hms2r { hms } {

    set h [hms2h $hms];
    set r [h2r $h];

    return $r;
}

# convert scalar julian date to besselian year
proc j2byear { j } {
    variable B1950;
    variable CB;

    set j [expr double($j)];	# enforce type

    set byear [expr (1950.0 + ($j - $B1950)*(100/$CB))];

    return $byear;
}

# convert scalar julian date to day of week (0 = Sunday)
proc j2dow { j } {
    set j [expr double($j)];	# enforce type

    # remember to round the JD to the next higher civil day
    set j [expr int($j + 0.5)];
    set dow [expr $j - 7 * int(($j + 1) / 7) + 1];

    return $dow;
}

# convert scalar julian date to gregorian proleptic calendar
proc j2gcal { j } {
    set j [expr double($j)];	# enforce type

    # remember to round the JD to the next higher civil day
    set l [expr int($j+0.5) + 68569];
    set n [expr int((4 * $l) / 146097)];
    set l [expr $l - int(((146097 * $n) + 3) / 4)];
    set i [expr int((4000 * ($l + 1)) / 1461101)];
    set l [expr $l - (int((1461 * $i) / 4) - 31)];
    set x [expr int((80 * $l) / 2447)];
    set d [expr $l - int((2447 * $x) / 80)];
    set l [expr int($x / 11)];
    set m [expr $x + 2 - (12 * $l)];
    set y [expr 100 * ($n - 49) + $i + $l];

    return [list $y $m $d];
}

# normalize scalar julian date
proc j2j { j } {
    set j [expr double($j)];	# enforce type

    # nothing to do here

    return $j;
}

# convert scalar julian date to julian proleptic calendar
proc j2jcal { j } {
    set j [expr double($j)];	# enforce type

    # remember to round the JD to the next higher civil day
    set x [expr int($j+0.5) + 1402];
    set k [expr int(($x - 1) / 1461)];
    set l [expr $x - (1461 * $k)];
    set n [expr int(($l - 1) / 365) - int($l / 1461)];
    set i [expr $l - (365 * $n) + 30];
    set x [expr int((80 * $i) / 2447)];
    set d [expr $i - int((2447 * $x) / 80)];
    set i [expr int($x / 11)];
    set m [expr $x + 2 - (12 * $i)];
    set y [expr (4 * $k) + $n + $i - 4716];

    return [list $y $m $d];
}

# convert scalar julian date to jd time
proc j2jd { j } {
    set j [expr double($j)];	# enforce type

    set jd [list $j {0.0 0.0 0.0}];

    return $jd;
}

# convert scalar julian date to julian year
proc j2jyear { j } {
    variable J2000;
    variable CJ;

    set j [expr double($j)];	# enforce type

    set jyear [expr (2000.0 + ($j - $J2000)*(100/$CJ))];

    return $jyear;
}

# convert scalar julian date to scalar year
proc j2y { j } {
    set j [expr double($j)];	# enforce type

    set jd [j2jd $j];
    set y [jd2y $jd];

    return $y;
}

# convert scalar julian date to ymd time
proc j2ymd { j } {
    set j [expr double($j)];	# enforce type

    set jd [j2jd $j];
    set ymd [jd2ymd $jd];

    return $ymd;
}

# convert gregorian calendar date to scalae julian date
proc jcal2j { y m d } {
    set y [expr int($y)];	# enforce type
    set m [expr int($m)];	# enforce type
    set d [expr int($d)];	# enforce type

    # normalize the month
    if {$m < 1} {
	set y [expr $y - (1-$m)/12];
	set m [expr ($m % 12) + 12];
    }
    if {$m > 12} {
	set y [expr $y + ($m-1)/12];
	set m [expr (($m-1) % 12) + 1];
    }

    #
    # In C, -7/12 is 0; in Tcl, it's -1!  Argh.
    # We will cobble up the C behavior
    #
    if {($m - 9) < 0} {
	set M [expr (9 - $m) / 7];
	set M [expr -1 * $M];
    } else {
	set M [expr ($m - 9) / 7];
    }

    set j [expr 367 * $y];
    set j [expr $j - (7 * ($y + 5001 + $M)) / 4];
    set j [expr $j + (275 * $m) / 9];	# m, not M, is correct
    set j [expr $j + $d];
    set j [expr $j + 1729777];

    return [expr double($j)];
}

# convert jd time to scalar julian date
proc jd2j { jd } {

    set j [lindex $jd 0];
    set hms [lindex $jd 1];
    set j [expr $j + [hms2h $hms] / 24.0];

    return $j;
}

# normalize jd time
proc jd2jd { jd } {

    # convert to decimal days
    set j [jd2j $jd];
    lset jd 0 [expr floor($j)];

    # promote the hours part
    set h [expr ($j - floor($j)) * 24.0];
    lset jd 1 0 [expr floor($h)];

    # promote the minutes part
    set m [expr ($h - floor($h)) * 60.0];
    lset jd 1 1 [expr floor($m)];

    # promote the seconds part
    set s [expr ($m - floor($m)) * 60.0];
    lset jd 1 2 $s;

    return $jd;
}

# convert jd time to scalar years
proc jd2y { jd } {

    set ymd [jd2ymd $jd];
    set y [ymd2y $ymd];

    return $y;
}

# convert jd time to ymd time
proc jd2ymd { jd } {

    set ymd {0 0 0.0 {0.0 0.0 0.0}};

    set j [lindex $jd 0];

    set gcal [j2gcal $j];
    lset ymd 0 [lindex $gcal 0];
    lset ymd 1 [lindex $gcal 1];
    lset ymd 2 [lindex $gcal 2];

    # get the fractional day
    set d [expr $j - floor($j)];

    # we do this next step because the rounding of j in j2gcal
    # either credits or debits us with 12 hours
    if {$d < 0.5} {
    	set d [expr $d + 0.5];
    } else {
    	set d [expr $d - 0.5];
    }
    lset ymd 2 [expr [lindex $ymd 2] + $d];

    # pick up the hours, minutes, seconds
    lset ymd 3 [lindex $jd 1];

    return $ymd;
}

# convert julian year to scalar julian date
proc jyear2j { jyear }  {
    variable J2000;
    variable CJ;

    set j [expr ($J2000 + ($jyear - 2000.0)*($CJ/100))];

    return $j;
}

# convert mjd time to ymd time
proc mjd2ymd { mjd } {
    variable MJD_0;

    set j [expr $mjd + $MJD_0];
    set ymd [j2ymd $j];

    return $ymd;
}

# convert radians to degrees
proc r2d { r } {
    set r [expr double($r)];	# enforce type

    set d [expr $r*(180/$math::constants::pi)];

    return $d;
}

# convert scalar radians to dms degrees
proc r2dms { r } {
    set r [expr double($r)];	# enforce type

    set d [r2d $r];
    set dms [d2dms $d];

    return $dms;
}

# convert scalar radians to scalar hours
proc r2h { r } {
    set r [expr double($r)];	# enforce type

    set h [expr $r*(12/$math::constants::pi)];

    return $h;
}

# convert scalar radians to hms hours
proc r2hms { r } {
    set r [expr double($r)];	# enforce type

    set h [r2h $r];
    set hms [h2hms $h];

    return $hms;
}

# normalize scalar radians to the range 0,2pi
proc r2r { r } {
    set r [expr double($r)];	# enforce type

    set twopi [expr 2*$math::constants::pi];

    if {$r < 0} {
	set r [expr $r + ceil($r / (-1*$twopi)) * $twopi];
    }
    if {$r >= $twopi} {
	set r [expr $r - floor($r / $twopi) * $twopi];
    }

    return $r;
}

# return the current time as scalar julian date
proc utc_now { } {
    # get the julian day number of the epoch
    set utc [expr [gcal2j 1970 1 1] - 0.5];

    # get the elapsed time since the unix epoch
    set t [clock seconds];

    # offset to now
    set utc [expr $utc + ($t / 86400.0)];

    return $utc;
}

# return the number of days in the given year
proc y2doy { y } {
    set y [expr int($y)];	# enforce an integer year

    if {($y % 400) == 0} {
    	set doy 366;
    } elseif {($y % 100) == 0} {
    	set doy 365;
    } elseif {($y % 4) == 0} {
    	set doy 366;
    } else {
    	set doy 365;
    }

    return $doy;
}

# convert scalar year to scalar julian date
proc y2j { y } {
    set y [expr double($y)];	# enforce type

    set jd [y2jd $y];
    set j [jd2j $jd];

    return $j;
}

# convert scalar year to jd time
proc y2jd { y } {
    set y [expr double($y)];	# enforce type

    set ymd [y2ymd $y];
    set jd [ymd2jd $ymd];

    return $jd;
}

# normalize scalar year
proc y2y { y } {
    set y [expr double($y)];	# enforce type

    # nothing to do

    return $y;
}

# convert scalar year to ymd time
proc y2ymd { y } {
    set y [expr double($y)];	# enforce type

    set f [expr floor($y)];
    set year [expr int($f)];
    set m 1;
    set d [expr ($y-$f) * [y2doy $year]];
    set ymd [list [expr int($y)] $m $d {0.0 0.0 0.0}];

    return $ymd;
}

# convert scalar year and day number to ymd time
proc ydd2ymd { y dd } {
    set y [expr int($y)];	# enforce type
    set dd [expr double($dd)];	# enforce type

    set ymd [list $y 1 $dd {0.0 0.0 0.0}];

    return $ymd;
}

# convert ymd time to scalar day of year
proc ymd2dd { ymd } {

    # normalize the time
    set ymd [ymd2ymd $ymd];
    set y [lindex $ymd 0];

    # get the julian date of the target date
    set j [ymd2j $ymd];

    # get the julian day number of the start of the eopch
    set j0 [expr [gcal2j $y 1 1] - 0.5];

    set dd [expr $j - $j0 + 1];

    return $dd;
}

# convert ymd time to scalar julian date
proc ymd2j { ymd } {

    set jd [ymd2jd $ymd];
    set j [jd2j $jd];

    return $j;
}

# convert ymd time to jd time
proc ymd2jd { ymd } {
    set y [expr int([lindex $ymd 0])];	# enforce type
    set m [expr int([lindex $ymd 1])];	# enforce type

    # normalize the month
    if {$m < 1} {
	set y [expr $y - (1-$m)/12];
	set m [expr ($m % 12) + 12];
    }
    if {$m > 12} {
	set y [expr $y + ($m-1)/12];
	set m [expr (($m-1) % 12) + 1];
    }

    set j [gcal2j $y $m 0];
    set j [expr $j + [lindex $ymd 2]];

    # julian days start 12 hours after civil days
    set j [expr $j - 0.5];

    # pick up the fractional day
    set hms [lindex $ymd 3];

    set jd [list $j $hms];

    return $jd;
}

# convert ymd time to mjd time
proc ymd2mjd { ymd } {
    variable MJD_0;

    set j [ymd2j $ymd];
    set mjd [expr $j - $MJD_0];

    return $mjd;
}

# convert ymd time to scalar year
proc ymd2y { ymd } {

    # normalize the time
    set ymd [ymd2ymd $ymd];

    set y [lindex $ymd 0];
    set dd [ymd2dd $ymd];
    set doy [y2doy $y];
    set y [expr $y + ($dd / $doy)];

    return $y;
}

# normalize ymd time
proc ymd2ymd { ymd } {
    set j [ymd2j $ymd];
    set gcal [j2gcal $j];
    set ymd {0 0 0.0 {0.0 0.0 0}};
    lset ymd 0 [lindex $gcal 0];
    lset ymd 1 [lindex $gcal 1];
    lset ymd 2 [lindex $gcal 2];

    set d [expr $j - floor($j)];

    # we do this next step because the rounding of j in j2gcal
    # either credits or debits us with 12 hours
    if {$d < 0.5} {
    	set d [expr $d + 0.5];
    } else {
    	set d [expr $d - 0.5];
    }

    # promote the hours
    set h [expr ($d - floor($d)) * 24.0];
    lset ymd 3 0 [expr floor($h)];

    # promote the minutes
    set m [expr ($h - floor($h)) * 60.0];
    lset ymd 3 1 [expr floor($m)];

    # promote the seconds
    set s [expr ($m - floor($m)) * 60.0];
    lset ymd 3 2 $s;

    return $ymd;
}

proc test { } {

    puts stdout "times::test:";

    set j 2451234.0;
    set y 2006;
    set m 7;
    set dd 40;
    set hh 30;
    set mm 30;
    set ss 30;
    set deg 70;
    set min 70;
    set sec 70;
    set r 1;
    set dms [list $deg $min $sec];
    set hms [list $hh $mm $ss];
    set ymd [list $y $m $dd $hms];
    set jd [list $j $hms];

    puts stdout "times::test: byear2j $y: [byear2j $y]";
    puts stdout "times::test: d2d $deg: [d2d $deg]";
    puts stdout "times::test: d2dms $deg: [d2dms $deg]";
    puts stdout "times::test: d2h $deg: [d2h $deg]";
    puts stdout "times::test: d2hms $deg: [d2hms $deg]";
    puts stdout "times::test: d2r $deg: [d2r $deg]";
    puts stdout "times::test: dms2d $dms: [dms2d $dms]";
    puts stdout "times::test: dms2dms $dms: [dms2dms $dms]";
    puts stdout "times::test: dms2h $dms: [dms2h $dms]";
    puts stdout "times::test: dms2hms $dms: [dms2hms $dms]";
    puts stdout "times::test: dms2r $dms: [dms2r $dms]";
    puts stdout "times::test: fmt_alpha $r: [fmt_alpha $r]";
    puts stdout "times::test: fmt_d $deg: [fmt_d $deg]";
    puts stdout "times::test: fmt_delta $r: [fmt_delta $r]";
    puts stdout "times::test: fmt_dms $dms: [fmt_dms $dms]";
    puts stdout "times::test: fmt_h $hh: [fmt_h $hh]";
    puts stdout "times::test: fmt_hms $hms: [fmt_hms $hms]";
    puts stdout "times::test: fmt_j $j: [fmt_j $j]";
    puts stdout "times::test: fmt_jd $jd: [fmt_jd $jd]";
    puts stdout "times::test: fmt_r $r: [fmt_r $r]";
    puts stdout "times::test: fmt_y $y: [fmt_y $y]";
    puts stdout "times::test: fmt_ymd $ymd: [fmt_ymd $ymd]";
    puts stdout "times::test: gcal2j $y $m $dd: [gcal2j $y $m $dd]";
    puts stdout "times::test: h2d $hh: [h2d $hh]";
    puts stdout "times::test: h2dms $hh: [h2dms $hh]";
    puts stdout "times::test: h2h $hh: [h2h $hh]";
    puts stdout "times::test: h2hms $hh: [h2hms $hh]";
    puts stdout "times::test: h2r $hh: [h2r $hh]";
    puts stdout "times::test: hms2d $hms: [hms2d $hms]";
    puts stdout "times::test: hms2dms $hms: [hms2dms $hms]";
    puts stdout "times::test: hms2h $hms: [hms2h $hms]";
    puts stdout "times::test: hms2hms $hms: [hms2hms $hms]";
    puts stdout "times::test: hms2r $hms: [hms2r $hms]";
    puts stdout "times::test: j2byear $j: [j2byear $j]";
    puts stdout "times::test: j2dow $j: [j2dow $j]";
    puts stdout "times::test: j2gcal $j:[j2gcal $j]";
    puts stdout "times::test: j2j $j: [j2j $j]";
    puts stdout "times::test: j2jcal $j: [j2jcal $j]";
    puts stdout "times::test: j2jd $j: [j2jd $j]";
    puts stdout "times::test: j2jyear $j: [j2jyear $j]";
    puts stdout "times::test: j2y $j: [j2y $j]";
    puts stdout "times::test: j2ymd $j: [j2ymd $j]";
    puts stdout "times::test: jcal2j $y $m $dd: [jcal2j $y $m $dd]";
    puts stdout "times::test: jd2j $jd: [jd2j $jd]";
    puts stdout "times::test: jd2jd $jd: [jd2jd $jd]";
    puts stdout "times::test: jd2y $jd: [jd2y $jd]";
    puts stdout "times::test: jd2ymd $jd: [jd2ymd $jd]";
    puts stdout "times::test: jyear2j $y: [jyear2j $y]";
    puts stdout "times::test: mjd2ymd 0: [mjd2ymd 0]";
    puts stdout "times::test: r2d $r: [r2d $r]";
    puts stdout "times::test: r2dms $r: [r2dms $r]";
    puts stdout "times::test: r2h $r: [r2h $r]";
    puts stdout "times::test: r2hms $r: [r2hms $r]";
    puts stdout "times::test: r2r $r: [r2r $r]";
    puts stdout "times::test: utc_now: [utc_now]";
    puts stdout "times::test: y2doy $y: [y2doy $y]";
    puts stdout "times::test: y2j $y:[y2j $y]";
    puts stdout "times::test: y2jd $y: [y2jd $y]";
    puts stdout "times::test: y2y $y: [y2y $y]";
    puts stdout "times::test: y2ymd $y: [y2ymd $y]";
    puts stdout "times::test: ydd2ymd $y $dd: [ydd2ymd $y $dd]";
    puts stdout "times::test: ymd2dd $ymd: [ymd2dd $ymd]";
    puts stdout "times::test: ymd2j $ymd: [ymd2j $ymd]";
    puts stdout "times::test: ymd2jd $ymd: [ymd2jd $ymd]";
    puts stdout "times::test: ymd2mjd $ymd: [ymd2mjd $ymd]";
    puts stdout "times::test: ymd2y $ymd: [ymd2y $ymd]";
    puts stdout "times::test: ymd2ymd $ymd: [ymd2ymd $ymd]";
}

# end of namespace
}
