# file: $RCSfile: jwp-vec.tcl,v $
# rcsid: $Id: jwp-vec.tcl,v 1.4 2006/08/07 21:26:27 jwp Exp $
# Copyright Jeffrey W Percival
# ######################################################################
# Space Astronomy Laboratory
# University of Wisconsin
# 1150 University Avenue
# Madison, WI 53706 USA
# ######################################################################
# Do not use this software without permission.
# Do not use this software without attribution.
# Do not remove or alter any of the lines above.
# ######################################################################

# ######################################################################
# $RCSfile: jwp-vec.tcl,v $ - Tcl implementation of JWP's libvec.a
# ######################################################################

# from tcllib (http://www.tcl.tk/software/tcllib/)
source constants.tcl
source linalg.tcl
package require math::constants
package require math::linearalgebra

namespace eval vec {

# Convert spherical coordinates into a cartesian vector.
# Right Ascension and Declination are in radians.
proc v3s2c { ra dec } {
    set v [math::linearalgebra::mkVector 3 0];

    #
    # tpm uses full-length, not unit, vectors;
    # for WHOMP targets we assume they are many parsecs away.
    #

    set R 1e9;

    set rcosd [expr $R * cos($dec)];

    set x [expr $rcosd * cos($ra)];
    set y [expr $rcosd * sin($ra)];
    set z [expr $R * sin($dec)];

    setVectorElement v 0 $x;
    setVectorElement v 1 $y;
    setVectorElement v 2 $z;

    return $v;
}

# Convert a cartesian vector to spherical angles.
proc v3c2s { v } {
    # default angles
    set ra 0;
    set dec 0;

    set x [getVectorElement $v 0];
    set y [getVectorElement $v 1];
    set z [getVectorElement $v 2];

    set R [expr sqrt($x*$x + $y*$y + $z*$z)];

    if {$R > 0} {
	if {$x == 0} {
	    if {$y < 0} {
		set ra [expr -1 * $math::constants::pi];
	    } elseif {$y > 0} {
		set ra [expr -1 * $math::constants::pi];
	    }
	} else {
	    set ra [expr atan2($y, $x)];
	}
    }

    set dec [expr atan2($z, sqrt($x*$x + $y*$y))];

    return [list $ra $dec];
}

#####################################
# augment the tcllib vector package #
#####################################

#
# Inexplicably, the tcllib linear algeba package
# does not include get and set procedures for vectors.
#

# Get a vector element.
proc getVectorElement { vector index } {
    set e [lindex $vector $index];

    return $e;
}

# Set a vector element.
# Side effect: the argument vector is changed.
proc setVectorElement { vector index value } {
    upvar $vector vec

    lset vec $index $value;

    return $vec;
}

#
# augment our matrix package
#

# A rotation about axis x.
proc m3Rx { x } {
    set c [expr cos($x)];
    set s [expr sin($x)];
    set r [math::linearalgebra::mkIdentity 3];
    math::linearalgebra::setelem r 1 1 $c;
    math::linearalgebra::setelem r 1 2 $s;
    math::linearalgebra::setelem r 2 1 [expr -1 * $s];
    math::linearalgebra::setelem r 2 2 $c;
    return $r;
}

# A rotation about axis y.
proc m3Ry { y } {
    set c [expr cos($y)];
    set s [expr sin($y)];
    set r [math::linearalgebra::mkIdentity 3];
    math::linearalgebra::setelem r 0 0 $c;
    math::linearalgebra::setelem r 0 2 [expr -1 * $s];
    math::linearalgebra::setelem r 2 0 $s;
    math::linearalgebra::setelem r 2 2 $c;
    return $r;
}

# A rotation about axis z.
proc m3Rz { z } {
    set c [expr cos($z)];
    set s [expr sin($z)];
    set r [math::linearalgebra::mkIdentity 3];
    math::linearalgebra::setelem r 0 0 $c;
    math::linearalgebra::setelem r 0 1 $s;
    math::linearalgebra::setelem r 1 0 [expr -1 * $s];
    math::linearalgebra::setelem r 1 1 $c;
    return $r;
}

proc test { } {

    puts stdout "vec::test:";

    set m [m3Rx $math::constants::pi/2];
    puts stdout "vec::test: m3Rx: m:\n[math::linearalgebra::show $m "%12.8f"]";
    set m [m3Ry $math::constants::pi/2];
    puts stdout "vec::test: m3Ry: m:\n[math::linearalgebra::show $m "%12.8f"]";
    set m [m3Rz $math::constants::pi/2];
    puts stdout "vec::test: m3Rz: m:\n[math::linearalgebra::show $m "%12.8f"]";
}

# end of namespace
}
