# file: $RCSfile: whomp-engine.tcl,v $
# rcsid: $Id: whomp-engine.tcl,v 1.8 2007/08/17 17:31:26 jwp Exp $
#
# whirc whomp
#
# computational engine for whomp planning
#

# Each exposure is taken at some offset from a target position.
# An offset is a pair of angles, represented as a list.
# A pattern is a list of offsets,
# and a map is a list of patterns.

# Comments on expanding dither and mapper patterns:
# The map is centered on the target position,
# which includes roll,
# so we represent the target position by an attitude matrix.
# The cells of the map are generated
# by performing attitude perturbations
# to the target, or reference, attitude.
#
# Dithers are centered on a map cell,
# and are generated exactly as we do map cells:
# perturbations to the cell attitude.

# ---------- pick up astro support
source jwp-astro.tcl;

# ---------- pick up the whomp constants
source whomp-constants.tcl;

namespace eval whomp {

# return true if the grid cell belongs to the shape
proc check_cell { shape ix iy nx ny } {
    set ok 1;	# assume it is ok

    # does this cell belong to the pattern?

    if {$shape == $whomp::PLUS} {
	if {$ix == int($nx/2)} {
	    # plus cell ok
	} elseif {$iy == int($ny/2)} {
	    # plus cell ok
	} else {
	    # plus cell skip
	    set ok 0;
	}
    } elseif {$shape == $whomp::XXXX} {
	if {$ix == $iy} {
	    # xxxx cell ok
	} elseif {($nx - 1 - $ix) == $iy} {
	    # xxxx cell ok
	} else {
	    # xxxx cell skip
	    set ok 0;
	}
    } else {
	# default is OK
    }

    return $ok;
}

# process a change in grid shape
proc new_grid_shape { shape nx ny nx_old ny_old } {

    #puts stdout "whomp::new_grid_shape: shape: $shape nx: $nx ny: $ny nx_old: $nx_old ny_old: $ny_old";
    if {$shape == $whomp::PLUS} {
	# any odd values
	if {$nx % 2 == 0} {
	    incr nx 1;
	}
	if {$ny % 2 == 0} {
	    incr ny 1;
	}
    } elseif {$shape == $whomp::XXXX} {
	# odd values that are equal
	if {$nx % 2 == 0} {
	    incr nx 1;
	}
	if {$ny % 2 == 0} {
	    incr ny 1;
	}
	# force them to be equal; take the larger
	if {$nx > $ny} {
	    set ny $nx;
	} else {
	    set nx $ny;
	}
    } else {
	# default is no change
    }

    #puts stdout "whomp::new_grid_shape: shape: $shape nx: $nx ny: $ny";
    set nx_old $nx;
    set ny_old $ny;

    return [list $nx $ny $nx_old $ny_old];
}

# process a change in grid size
# for plus and cross,
# force the next higher odd value
proc new_grid_size { shape nx ny nx_old ny_old } {

    #puts stdout "whomp::new_grid_size: shape: $shape nx: $nx ny: $ny nx_old: $nx_old ny_old: $ny_old";

    if {$shape == $whomp::PLUS} {
	# any odd values
	if {$nx % 2 == 0} {
	    if {$nx < $nx_old} {
		incr nx -1;
	    } else {
		incr nx 1;
	    }
	}
	if {$ny % 2 == 0} {
	    if {$ny < $ny_old} {
		incr ny -1;
	    } else {
		incr ny 1;
	    }
	}
    } elseif {$shape == $whomp::XXXX} {
	if {$nx % 2 == 0} {
	    if {$nx < $nx_old} {
		incr nx -1;
	    } else {
		incr nx 1;
	    }
	    set ny $nx;
	}
	if {$ny % 2 == 0} {
	    if {$ny < $ny_old} {
		incr ny -1;
	    } else {
		incr ny 1;
	    }
	    set nx $ny;
	}
    } else {
	# any values are allowed
    }
    set nx_old $nx;
    set ny_old $ny;

    #puts stdout "whomp::new_grid_size: shape: $shape nx: $nx ny: $ny nx_old: $nx_old ny_old: $ny_old";

    return [list $nx $ny $nx_old $ny_old];
}

# return the number of locations in the pattern
proc pattern_count { pattern } {

    #puts stdout "whomp::pattern_count: pattern: $pattern";

    # break up the list

    set shape [lindex $pattern 0];
    set nx [lindex $pattern 1];
    set ny [lindex $pattern 2];

    set count 0;

    # loop through the pattern, counting legal cells

    for {set iy 0} {$iy < $ny} {incr iy} {
	for {set ix 0} {$ix < $nx} {incr ix} {
	    set ok [whomp::check_cell $shape $ix $iy $nx $ny];
	    if {$ok} {
		incr count;
	    }
	}
    }

    return $count;
}

# compute the pattern size in arcseconds
proc pattern_size { pattern } {

    #puts stdout "whomp::pattern_size: pattern: $pattern";

    set shape [lindex $pattern 0];
    set nx [lindex $pattern 1];
    set ny [lindex $pattern 2];
    set dx [lindex $pattern 3];
    set dy [lindex $pattern 4];

    set w [expr $whomp::virgo_w + (($nx-1) * $dx)];
    set h [expr $whomp::virgo_h + (($ny-1) * $dy)];

    # allow a margin of half a chip around the pattern
    set w [expr $whomp::virgo_w/2 + $w + $whomp::virgo_w/2];
    set h [expr $whomp::virgo_h/2 + $h + $whomp::virgo_h/2];

    #puts stdout "whomp::pattern_size: w: $w h: $h";

    # choose the larger
    if {$w > $h} {
	set size $w;
    } else {
	set size $h;
    }

    return $size;
}

# return the map size in arcseconds
proc map_size { mpattern dpattern } {

    set msize [pattern_size $mpattern];
    set dsize [pattern_size $dpattern];
    #puts stdout "whomp::map_size: msize: $msize dsize: $dsize";

    set size [expr $msize + ($dsize/2)];

    return $size;
}

# compute the canvas scale in pixels per arcsecond
proc canvas_scale { mpattern dpattern canvas_w canvas_h } {

    # get the map size in arcseconds

    set size [map_size $mpattern $dpattern];
    #puts stdout "whomp::canvas_scale: size: $size";

    # compute the scale in each dimension

    set scale_w [expr double($canvas_w) / $size];
    set scale_h [expr double($canvas_h) / $size];
    #puts stdout "whomp::canvas_scale: scale w: $scale_w h: $scale_h";

    # choose the smaller, and use it in both dimensions
    if {$scale_h < $scale_w} {
	set scale $scale_h;
    } else {
	set scale $scale_w;
    }
    #puts stdout "whomp::canvas_scale: scale: $scale";

    return $scale;
}

# expand a pattern into a list of angular offsets.
# we allow two offsets that we fold into each member of the pattern;
# when generating a dither pattern within a map pattern,
# the offset gives the "base position" for the dither.
# the base position would vary with each map cell.
proc expand_pattern { pattern x0 y0 } {

    #puts stdout "whomp::expand_pattern: pattern: $pattern x0: $x0 y0: $y0";

    # break up the list

    set shape [lindex $pattern 0];
    set nx [lindex $pattern 1];
    set ny [lindex $pattern 2];
    set dx [lindex $pattern 3];
    set dy [lindex $pattern 4];

    set plist {};	# the list of offsets

    # loop through the pattern, processing legal cells

    for {set iy 0} {$iy < $ny} {incr iy} {
	for {set ix 0} {$ix < $nx} {incr ix} {
	    set ok [whomp::check_cell $shape $ix $iy $nx $ny];
	    if {$ok} {
		# we measure the spacing in half-offsets
		set xsteps [expr 2*$ix - ($nx-1)];
		set ysteps [expr ($ny-1) - 2*$iy];
		set xoffset [expr $xsteps * ($dx/2.0)];
		set yoffset [expr $ysteps * ($dy/2.0)];
		#puts stdout "whomp::expand_pattern: ix: $ix iy: $iy xsteps: $xsteps ysteps: $ysteps xoffset $xoffset yoffset $yoffset";
		# add in the provided offset
		set xoffset [expr $xoffset + $x0];
		set yoffset [expr $yoffset + $y0];

		lappend plist [list $xoffset $yoffset];
	    }
	}
    }

    return $plist;
}

# expand a map into a list of dithers
proc expand_map { mpattern dpattern } {

    # expand the mapper pattern

    set plist [expand_pattern $mpattern 0 0];
    #puts stdout "whomp::expand_map: plist: $plist";

    # for each offset in the pattern, replace it with a list of dithers

    set mlist {};
    foreach p $plist {
	#puts stdout "whomp::expand_map: p: $p";
	set x0 [lindex $p 0];
	set y0 [lindex $p 1];
	set dlist [expand_pattern $dpattern $x0 $y0];
	#puts stdout "whomp::expand_map: dlist: $dlist";
	lappend mlist $dlist;
    }

    return $mlist;
}

# convert a map into a list of attitude matrix lists
proc map2matrix { mpattern dpattern } {

    set mlist [expand_map $mpattern $dpattern];

    set alist {};

    foreach dlist $mlist {
	puts stdout "whomp::map2matrix: dlist: $dlist";
	set foo {};
	foreach d $dlist {
	    puts stdout "whomp::map2matrix: d: $d";
	    set x [lindex $d 0];
	    set y [lindex $d 1];

	    # convert the offsets into rotation matrices
	    set ry [vec::m3Ry [times::as2r $x]];
	    puts stdout "whomp::map2matrix: ry:\n[math::linearalgebra::show $ry "%12.8f"]";
	    set rx [vec::m3Rx [times::as2r $y]];
	    puts stdout "whomp::map2matrix: rx:\n[math::linearalgebra::show $rx "%12.8f"]";
	    # initialize
	    set m [math::linearalgebra::mkIdentity 3];
	    puts stdout "whomp::map2matrix: m:\n[math::linearalgebra::show $m "%12.8f"]";
	    # now move to the current position in the pattern
	    set m [math::linearalgebra::matmul $ry $m];
	    puts stdout "whomp::map2matrix: m:\n[math::linearalgebra::show $m "%12.8f"]";
	    set m [math::linearalgebra::matmul $rx $m];
	    puts stdout "whomp::map2matrix: m:\n[math::linearalgebra::show $m "%12.8f"]";
	    lappend foo $m;
	}
	lappend alist $foo;
    }

    return $alist;
}

# convert a map into a list of boxes in canvas units
proc map2boxlist { mpattern dpattern canvas_w canvas_h } {

    # get the scale factor

    set scale [canvas_scale $mpattern $dpattern $canvas_w $canvas_h];
    #puts stdout "whomp::map2boxlist: scale: $scale";

    # compute the box size in pixels

    set box_w [expr int($scale * $whomp::virgo_w)];
    set box_h [expr int($scale * $whomp::virgo_h)];
    #puts stdout "whomp::map2boxlist: box w: $box_w h: $box_h";

    # build the list of lists
    set mlist [expand_map $mpattern $dpattern];
    #puts stdout "whomp::map2boxlist: mlist: $mlist";

    set boxlist {};
    foreach dlist $mlist {
	#puts stdout "whomp::map2boxlist: dlist: $dlist";
	set foo {};
	foreach d $dlist {
	    #puts stdout "whomp::map2boxlist: d: $d";

	    # get the center of this cell in arcseconds w.r.t center of canvas

	    set offset_w [lindex $d 0];
	    set offset_h [lindex $d 1];
	    #puts stdout "dither::redraw: offset w: $offset_w h: $offset_h";

	    # find the center of this cell in pixels w.r.t corner of canvas

	    set center_w [expr $canvas_w/2 + ($scale * $offset_w)];
	    set center_h [expr $canvas_h/2 + ($scale * $offset_h)];
	    #puts stdout "dither::redraw: center w: $center_w h: $center_h";

	    # compute the corners of the box

	    set x1 [expr int($center_w - $box_w/2)];
	    set y1 [expr int($center_h - $box_h/2)];
	    set x2 [expr int($center_w + $box_w/2)];
	    set y2 [expr int($center_h + $box_h/2)];
	    #puts stdout "rectangle $x1 $y1 $x2 $y2";

	    set box [list $x1 $y1 $x2 $y2];

	    lappend foo $box;
	}

	lappend boxlist $foo;
    }

    return $boxlist;
}

proc test { } {

    # set up a grid change

    set grid [list $whomp::PLUS 2 2 3 3];
    puts stdout "whomp::test: grid: $grid";
    puts stdout "whomp::test: check_cell $grid: [eval check_cell $grid]";
    puts stdout "whomp::test: new_grid_shape $grid: [eval new_grid_shape $grid]";
    puts stdout "whomp::test: new_grid_size $grid: [eval new_grid_size $grid]";

    # create mapper and dither patterns

    set mpat [list $whomp::PLUS 3 3 200 200];
    set dpat [list $whomp::PLUS 3 3 20 20];
    puts stdout "whomp::test: mpat: $mpat size: [pattern_size $mpat]";
    puts stdout "whomp::test: dpat: $dpat size: [pattern_size $dpat]";
    puts stdout "whomp::test: mpat: $mpat dpat: $dpat scale: [canvas_scale $mpat $dpat 640 480]";

    set map [expand_map $mpat $dpat];
    puts stdout "whomp::test: map: $map";
    foreach dlist $map {
	puts stdout "whomp::test: dlist: $dlist";
	foreach d $dlist {
	    puts stdout "whomp::test: d: $d";
	}
    }

    set alist [map2matrix $mpat $dpat];
    puts stdout "whomp::test: alist: $alist";
    foreach dlist $alist {
	puts stdout "whomp::test: dlist: $dlist";
	foreach d $dlist {
	    puts stdout "whomp::test: d: $d";
	}
    }

    set boxlist [map2boxlist $mpat $dpat 640 480];
    puts stdout "whomp::test: boxlist: $boxlist";
    foreach dlist $boxlist {
	puts stdout "whomp::test: dlist: $dlist";
	foreach d $dlist {
	    puts stdout "whomp::test: d: $d";
	}
    }

    return;
}

# end of namespace
}
